/*
 * Copyright (c) 2022-2025 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *                                         Gianluca Corsini on Fri Sep 02 2022
 */
#include "acdxsim.h"

#include <sys/time.h>
#include <err.h>
#include <math.h>
#include <pthread.h>

#include <iostream>
#include <string>
#include <vector>

#include "gz/math/Rand.hh"
#include "gz/plugin/Register.hh"
#include "gz/sim/Joint.hh"
#include "gz/sim/System.hh"
#include "gz/sim/Util.hh"
#include "sdf/Element.hh"

#include "libdynamixel.h"

using namespace gz;

#define DXSIM_DEFAULT_PTY		"/tmp/pty-dynamixel"
#define DXSIM_DEFAULT_MOTOR_MODEL	"MX64T"
#define DXSIM_SUPPORTED_MOTOR_MODELS	"'MX64T', 'MX28T'"


class dxsim :
  public sim::System,
  public sim::ISystemConfigure,
  public sim::ISystemPreUpdate,
  public sim::ISystemPostUpdate
{
  struct motor {
    sim::Joint joint;
    int spin;
    struct dynamixel_s *dynamixel;
    double noise;
    struct dynamixelstate_motor_s current;
    double cmd, err[3];
  };

  std::vector<struct motor> motors;
  struct dynamixel_pty_s *pty;
  bool alive;

  pthread_t reader;
  pthread_mutex_t dxsim_mutex = PTHREAD_MUTEX_INITIALIZER;

 public:

  /* --- Configure --------------------------------------------------------- */

  void
  Configure(const sim::Entity &entity, const sdf::ElementConstPtr &sdf,
            sim::EntityComponentManager &ecm, sim::EventManager &)
  {
    std::unordered_set<sim::Entity> entities;
    std::string j_name, pty_name, model_name, spin, eeprom;
    sdf::ElementConstPtr m, j, s;
    enum dx_model model_id;
    double p0, v0;
    motor motor;
    int id;

    /* pty */
    sdf->Get<std::string>("pty", pty_name, DXSIM_DEFAULT_PTY);

    pty = dynamixel_init_pty();
    if (!pty)
      gzerr << "[dxsim] cannot initialize pty\n";
    else if (dynamixel_open_pty(pty, pty_name.c_str()))
      gzerr << "[dxsim] cannot open pty " << pty_name << "\n";
    else
      warnx("dxsim created pty `%s'", pty_name.c_str());

    /* motors */
    if (!sdf->HasElement("motors")) {
      gzerr << "[dxsim] missing <motors> tag\n";
      return;
    }

    m = sdf->FindElement("motors");
    for(m = m->FindElement("motor"); m; m = m->GetNextElement("motor")) {
      /* joint */
      j = m->FindElement("joint");
      if (!j) {
        gzerr << "[dxsim] ignoring <motor> without <joint> tag\n";
        continue;
      }
      j_name = j->GetValue()->GetAsString();
      entities = sim::entitiesFromScopedName(j_name, ecm, entity);
      if (entities.empty()) {
        gzerr << "[dxsim] ignoring non-existent joint '" << j_name << "'\n";
        continue;
      }
      motor.joint = sim::Joint(*entities.begin());
      if (!motor.joint.Valid(ecm)) {
        gzerr << "[dxsim] ignoring invalid joint '" << j_name << "'\n";
        continue;
      }
      if (motor.joint.Type(ecm).
          value_or(sdf::JointType::INVALID) != sdf::JointType::REVOLUTE) {
        gzerr << "[dxsim] ignoring joint '" << j_name
              << "' not of the hinge type\n";
        continue;
      }

      /* joint attribute spin */
      if (j->HasAttribute("spin")) {
        spin = j->GetAttribute("spin")->GetAsString();
        if (spin == "cw") motor.spin = -1;
        else if (spin == "ccw") motor.spin = 1;
        else {
          gzerr << "[dxsim] ignoring invalid spin '" << spin
                << "' (must be cw or ccw)\n";
          motor.spin = 1;
        }
      } else
        motor.spin = 1;

      /* initial condition */
      m->Get<double>("p0", p0, 0.);
      m->Get<double>("v0", v0, 0.);

      motor.joint.ResetPosition(ecm, {p0});
      motor.joint.ResetVelocity(ecm, {v0});

      /* read effective position, velocity and torque */
      motor.joint.EnablePositionCheck(ecm);
      motor.joint.EnableVelocityCheck(ecm);
      motor.joint.EnableTransmittedWrenchCheck(ecm);

      /* arbitrary limits */
      motor.joint.SetVelocityLimits(ecm, { math::Vector2d{-25., 25.} });
      motor.joint.SetEffortLimits(ecm, { math::Vector2d{-10., 10.} });

      /* model */
      m->Get<std::string>("model", model_name, "");
      if (model_name == "") {
        gzmsg << "[dxsim] motor model not specified, using default `"
              << DXSIM_DEFAULT_MOTOR_MODEL << "'\n";
        model_name = DXSIM_DEFAULT_MOTOR_MODEL;
      }
      if (model_name == "MX28T")
        model_id = DX_MX28T;
      else if (model_name == "MX64T")
        model_id = DX_MX64T;
      else {
        gzerr << "[dxsim] ignoring unsupported motor model `"
              << model_name << "'\n";
        gzerr << "supported models are "
              << DXSIM_SUPPORTED_MOTOR_MODELS << "\n";
        continue;
      }

      /* eeprom */
      m->Get<std::string>("eeprom", eeprom, DX_DEFAULT_EEPROM);

      /* id */
      m->Get<int>("id", id, DX_DEFAULT_ID);
      if (id < 0 || id > 255) {
        gzerr << "[dxsim] invalid motor id `" << id << "'\n";
        continue;
      }
      if (id == DX_BROADCAST_ID) {
        gzerr << "[dxsim] invalid motor id `" << id << "'\n";
        continue;
      }

      /* noise */
      m->Get<double>("noise", motor.noise, 0.);
      motor.noise = fabs(motor.noise);

      /* init dynamixel */
      motor.dynamixel = dynamixel_init(model_id, id, eeprom.c_str());
      if (!motor.dynamixel) {
        gzerr << "[dxsim] could not initialize motor id " << id << "\n";
        continue;
      }

      /* init current and desired motor data */
      motor.current = {
        .effort   = 0.,
        .velocity = v0,
        .position = p0
      };
      motor.cmd = motor.err[0] = motor.err[1] = motor.err[2] = 0.;

      motors.push_back(motor);
      warnx("configured motor %10.10s, id %3u, eeprom %s",
            model_name.c_str(),
            dynamixel_get_id(motor.dynamixel),
            dynamixel_get_eeprom_path(motor.dynamixel));
    }

    /* create reader thread */
    if (sdf->HasElement("period"))
      gzerr << "[dxsim] ignoring unused period tag\n";
    alive = true;
    pthread_create(&reader, NULL, reader_thread, this);
  }


  /* --- PreUpdate --------------------------------------------------------- */

  void
  PreUpdate(const sim::UpdateInfo &, sim::EntityComponentManager &ecm)
  {
    const struct dynamixelstate_motor_s *desired;
    double noise;
    motor *m;

    pthread_mutex_lock(&dxsim_mutex);
    for (uint8_t i = 0; i < motors.size(); i++) {
      m = &motors[i];

      noise =  m->noise > 0. ? math::Rand::DblNormal(0., m->noise) : 0.;
      desired = dynamixel_get_motor_ref(m->dynamixel);

      m->err[2] = m->err[1];
      m->err[1] = m->err[0];
      m->err[0] = 0.;

      switch (dynamixel_get_op_mode(m->dynamixel)) {
        case DX_POSITION_MODE:
        case DX_EXT_POSITION_MODE: {
          const struct pidff_s *pid =
            dynamixel_get_position_pidff(m->dynamixel);

          m->err[0] = m->spin * desired->position - m->current.position;
          m->cmd +=
            (pid->kP + pid->kI + pid->kD) * m->err[0]
            - (pid->kP + 2 * pid->kD) * m->err[1]
            + pid->kD * m->err[2];

          m->joint.SetForce(ecm, { m->cmd });
          break;
        }

        case DX_VELOCITY_MODE: {
          const struct pi_s *pid =
            dynamixel_get_velocity_pi(m->dynamixel);

          m->err[0] = m->spin * desired->velocity - m->current.velocity;
          m->cmd +=
            (pid->kP + pid->kI) * m->err[0]
            - pid->kP * m->err[1];

          m->joint.SetForce(ecm, { m->cmd });
          break;
        }

        case DX_TORQUE_MODE:
          m->cmd = m->spin * (desired->effort + noise);
          m->joint.SetForce( ecm, { m->cmd });
          break;

        case DX_TORQUE_POSITION_MODE:
          break;

        default:
          gzerr << "[dxsim] undefined operating mode "
                << dynamixel_get_op_mode(motors[i].dynamixel) << "\n";
          break;
      }
    }
    pthread_mutex_unlock(&dxsim_mutex);
  }


  /* --- PostUpdate -------------------------------------------------------- */

  void
  PostUpdate(const sim::UpdateInfo &,
             const sim::EntityComponentManager &ecm)
  {
    motor *m;
    size_t i;

    pthread_mutex_lock(&dxsim_mutex);
    for (i = 0; i < motors.size(); i++) {
      m = &motors[i];

      auto W = motors[i].joint.TransmittedWrench(ecm);
      if (W.has_value()) {
        auto t = W.value()[0].torque();

        m->current.effort =
          math::Vector3d{t.x(), t.y(), t.z()}
          .Dot(m->joint.Axis(ecm).value()[0].Xyz());
      } else
        m->current.effort = 0.;

      m->current.velocity =
        m->joint.Velocity(ecm).value_or(std::vector<double>{0.})[0];

      m->current.position =
        m->joint.Position(ecm).value_or(std::vector<double>{0.})[0];

      dynamixel_set_motor_state(m->dynamixel, &m->current);
    }
    pthread_mutex_unlock(&dxsim_mutex);
  }


  /* --- ~dxsim ------------------------------------------------------------ */

  ~dxsim()
  {
    size_t i;

    pthread_mutex_lock(&dxsim_mutex);
    alive = false;
    for (i = 0; i < motors.size(); i++)
      dynamixel_fini(motors[i].dynamixel);
    pthread_mutex_unlock(&dxsim_mutex);
    pthread_join(reader, NULL);
    dynamixel_fini_pty(pty);
  }


 private:

  static void *
  reader_thread(void *data)
  {
    dxsim *plugin = (dxsim *)data;
    motor *motor;
    size_t i;
    int s;

    pthread_mutex_lock(&plugin->dxsim_mutex);
    while(plugin->alive) {
      pthread_mutex_unlock(&plugin->dxsim_mutex);

      s = dynamixel_wait_msg(plugin->pty, 500);

      pthread_mutex_lock(&plugin->dxsim_mutex);
      if (!plugin->alive) break;

      if (s == 0) continue; /* timeout */
      else if (s < 0) {
        gzerr << "[dxsim] error when reading "
              << dynamixel_pty_name(plugin->pty) << "\n";
        continue;
      }

      while((s = dynamixel_get_msg(plugin->pty)) == 1) {
        for (i = 0; i < plugin->motors.size(); i++) {
          motor = &plugin->motors[i];
          if (motor->dynamixel)
            s = dynamixel_process_msg(motor->dynamixel, plugin->pty);
          else
            gzerr << "[dxsim] could not access dynamixel id "
                  << dynamixel_get_id(motor->dynamixel) << "\n";
        }
      }
      if (s < 0)
        gzerr << "[dxsim] cannot read message from %s"
              << dynamixel_pty_name(plugin->pty) << "\n";
    }
    pthread_mutex_unlock(&plugin->dxsim_mutex);

    warnx("dxsim disconnecting");
    return NULL;
  }
};

/* register plugin */
GZ_ADD_PLUGIN(
  dxsim,
  sim::System,
  dxsim::ISystemConfigure,
  dxsim::ISystemPreUpdate,
  dxsim::ISystemPostUpdate)
